package client;

import java.net.*;
import java.io.IOException;
import java.util.HashMap;

/**
 * The instant messenger application -- use it to communicate with
 * your friends over the internet! :-)
 */
public class InstantMessenger {
	/* THE InstantMessenger */
	public static InstantMessenger IM = null;

	/* For convenience we keep a map id --> User */
	private HashMap map = new HashMap();

	/* the GUI */
	private IMFrame gui;

	/* network stuff */
	private DatagramSocket socket; // UDP socket
	private MessageReceiver receiver;
	private MessageSender sender;
	private RegistrationClient registrationClient;

	/* some variables */
	private String name;

	/**
	 * Creates an instant messenger that connects (TCP) to a
	 * registration server at the specified address and makes the user
	 * to be known under the given name and port.
	 */
	public InstantMessenger(InetAddress serverAddress, String name, int port) {
		this.name = name;

		/* set-up socket-using facilities */
		try {
			/* Create UDP socket */
			socket = new DatagramSocket(port);
			/* Create receiver/sender threads */
			receiver = new MessageReceiver(socket);
			sender = new MessageSender(socket);
			/* Create the RegistrationClient, that registers with the server */
			registrationClient = new RegistrationClient(serverAddress, name, port);
		} catch (SocketException se) {
			System.out.println("Cannot send/receive messages, exiting.");
			se.printStackTrace();
			System.exit(0);
		}

		/* set-up the gui */
		gui = new IMFrame(sender, name);
		gui.pack();
	}

	/**
	 * "Destructor"
	 */
	protected void finalize() throws IOException {
		socket.close();
	}

	/**
	 * Start registering ourselves and receiving messages.
	 */
	void start() {
		/* Now start the receiver/sender threads */
		receiver.start();
		registrationClient.start();

		/* And show the gui */
		gui.setVisible(true);
	}

	// ----- User these methods, if necessary --------------------------------------

	/**
	 * Returns the user with the given id, _null_ otherwise. Use this
	 * message to get information about a user (name).
	 */
	User getUser(byte[] ip, int port) {
		return (User)map.get(createHash(ip, port));
	}

	/**
	 * Updates our view of the users. Use this method after receiving
	 * a USERLIST messages to make the new users available to the IM.
	 */
	synchronized void setUsers(User[] users) {
		/* Putting the users into our HashMap */
		for (int i = 0; i < users.length; i++) {
			User user = users[i];
			map.put(createHash(user.getIP(), user.getPort()), user);
		}

		/* display it in the GUI */
		gui.setUserList(users);
	}

	/**
	 * Displays a message that we received from another user. Use this
	 * message after receiving a MESSAGE message to show the message in
	 * the gui.
	 */
	void displayMessage(String message) {
		if (!message.startsWith(name + ": ")) {
			/* delegate to the GUI */
			gui.displayUserMessage(message);
		}
	}

	/**
	 * Calculates a hash value from the given ip address and port (used for
	 * the hashmap).
	 */
	private Long createHash(byte[] ip, int port) {
		long tmp = (long) (((ip[0] & 0xff) << 40) | ((ip[1] & 0xff) << 32) | ((ip[2] & 0xff) << 24) | ((ip[3] & 0xff) << 16)) + port;
		return new Long(tmp);
	}

	// -----------------------------------------------------------------------------

	/**
	 * Invoked by the jvm to start the program.
	 */
	public static void main(String[] arguments) {
		if (arguments.length != 3) {
			printUsage();
		} else {
			try {
				/* Parsing the command line parameters */
				InetAddress address = InetAddress.getByName(arguments[0]);
				String name = arguments[1];
				if (name.length() > 255) {
					printUsage();
				}
				int port = Integer.parseInt(arguments[2]);

				/* Starting the client */
				IM = new InstantMessenger(address, name, port);
				IM.start();
			} catch (Exception e) {
				System.out.println("Ooops, something went wrong.");
				e.printStackTrace();
				System.out.println();
				printUsage();
			}
		}
	}

	/**
	 * Prints some information about the usage of this program/its arguments.
	 */
	private static void printUsage() {
		System.out.println("Usage: java InstantMessenger <server-address> <name> <port>");
		System.out.println("Where");
		System.out.println("<server-address> is either 'dcg.ethz.ch' (Java) or 'byteslave.ethz.ch' (C) - try both!,");
		System.out.println("<name> is your group (eg. IFW_A34) and real name(s) (max. 255 characters), and");
		System.out.println("<port> is a free port number on your machine > 1023, eg. '6789'");
		System.exit(0);
	}
}